' Copyright 2008 - 2010 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Imports System.IO
Imports System.Reflection

''' <summary>
''' TileManager logic
''' </summary>
Public Class TileManager
    Private myTileProviders As New TileProviders
    Private myZoom As Integer
    Private myLatitude As Double
    Private myLongitude As Double
    Private myTileProviderList As New List(Of String)
    Public Shared myCultureInfo As New System.Globalization.CultureInfo("en-US")
    Public Sub New()
        Dim oConfig As New XMLConfig(GetAppPath() & IO.Path.DirectorySeparatorChar & "config.xml")
        myZoom = oConfig.GetIntegerSetting("LastZoom", 13)
        myLatitude = oConfig.GetDoubleSetting("LastLatitude", 0)
        myLongitude = oConfig.GetDoubleSetting("LastLongitude", 0)

        ' Smoothly mirgrate "old" settings
        If oConfig.GetSetting("TileStorageFolder") <> "" Then
            oConfig.SetSetting("TileProviderName1", "imported provider")
            oConfig.SetSetting("TileStorageFolder1", oConfig.GetSetting("TileStorageFolder"))
            oConfig.DeleteSetting("TileStorageFolder")
        End If
        If oConfig.GetSetting("CurrentTileServerUrl") <> "" Then
            oConfig.SetSetting("TileServerUrl1", oConfig.GetSetting("CurrentTileServerUrl"))
            oConfig.DeleteSetting("CurrentTileServerUrl")
        End If
        Dim tileServerFound As Boolean = True
        Dim i As Integer = 1
        While tileServerFound
            Dim strTileProviderName As String = oConfig.GetSetting("TileProviderName" & i)
            If strTileProviderName.Length <> 0 Then
                Dim strTileFolder As String = oConfig.GetSetting("TileStorageFolder" & i)
                If IO.Directory.Exists(strTileFolder) Then
                    If myTileProviders.AddTileProvider(strTileProviderName, strTileFolder, oConfig.GetSetting("TileServerUrl" & i), oConfig.GetIntegerSetting("TileServerMinZoom" & i, MIN_ZOOM), oConfig.GetIntegerSetting("TileServerMaxZoom" & i, MAX_ZOOM)) Then
                        myTileProviderList.Add(strTileProviderName)
                    End If
                End If
            Else
                tileServerFound = False
            End If
            i += 1
        End While
    End Sub
    Public Function GetAppPath() As String
        GetAppPath = Path.GetDirectoryName([Assembly].GetExecutingAssembly().GetModules(0).FullyQualifiedName)
    End Function
    ''' <summary>
    ''' Delete tiles which are older than beforeDate.
    ''' </summary>
    ''' <param name="beforeDate">The date for tiles to be handled.</param>
    Public Sub DeleteOldTiles(ByVal beforeDate As Date)
        If MsgBox("Are you sure that you want to delete all old tiles?", MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
            System.Windows.Forms.Cursor.Current = Cursors.WaitCursor
            Dim useWalkerAction As WalkerAction_DeleteOld = New WalkerAction_DeleteOld(beforeDate)
            walkDirectories(myTileProviders.GetCurrentFolderName, useWalkerAction)
            System.Windows.Forms.Cursor.Current = Cursors.Default
            MsgBox("finished. Deleted tiles: " & useWalkerAction.GetDeletedTilesCount.ToString, MsgBoxStyle.Information)
        End If
    End Sub
    ''' <summary>
    ''' Update tiles which are older than beforeDate.
    ''' </summary>
    ''' <param name="beforeDate">The date for tiles to be handled.</param>
    Public Sub UpdateOldTiles(ByVal beforeDate As Date)
        If myTileProviders.GetCurrentURL = "" Then
            MsgBox("Set up the tile server in OSMtracker first.", MsgBoxStyle.Information)
        Else
            If MsgBox("Are you sure that you want to update all tiles? This might take a long time", MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
                System.Windows.Forms.Cursor.Current = Cursors.WaitCursor
                Dim useWalkerAction As WalkerAction_Update = New WalkerAction_Update(myTileProviders.GetCurrentURL, beforeDate)
                walkDirectories(myTileProviders.GetCurrentFolderName, useWalkerAction)
                System.Windows.Forms.Cursor.Current = Cursors.Default
                MsgBox("finished. Updated tiles: " & useWalkerAction.GetUpdatedTilesCount.ToString, MsgBoxStyle.Information)
            End If
        End If
    End Sub
    Private Sub walkDirectories(ByVal rootDirectory As String, ByVal action As WalkerAction)
        Dim progress As New Progress
        progress.Show()
        progress.BringToFront()
        directoryWalker(rootDirectory, action, progress)
        progress.Close()
    End Sub
    Private Sub directoryWalker(ByVal rootDirectory As String, ByVal action As WalkerAction, ByVal progress As Progress)
        Dim directories As String() = Directory.GetDirectories(rootDirectory)
        progress.IncreaseDepth(directories.Length)
        For Each subDir As String In directories
            directoryWalker(subDir, action, progress)
        Next
        If progress.IsAborted Then
            Exit Sub
        End If
        For Each file As String In Directory.GetFiles(rootDirectory, "*.png")
            action.DoAction(file)
        Next
        If directories.Length = 0 Then
            progress.Increase()
        End If
        progress.DecreaseDepth()
    End Sub
    ''' <summary>
    ''' Displays information about the selected tile folder.
    ''' </summary>
    ''' <param name="beforeDate">The date for tiles to be reported as old.</param>
    Public Sub GetTileInfos(ByVal beforeDate As Date)
        System.Windows.Forms.Cursor.Current = Cursors.WaitCursor
        Dim useWalkerAction As WalkerAction_Info = New WalkerAction_Info(beforeDate)
        walkDirectories(myTileProviders.GetCurrentFolderName, useWalkerAction)
        System.Windows.Forms.Cursor.Current = Cursors.Default
        MsgBox("Number of all tiles: " & useWalkerAction.GetNumberOfTiles.ToString & vbCrLf & _
               "Size of all tiles in KiB: " & (useWalkerAction.GetSizeOfAllTiles / 1024).ToString("0.000") & vbCrLf & _
               "Number of tiles by date filter: " & useWalkerAction.GetAffectedTilesCount & vbCrLf & _
               "Size of tiles  by date filter in KiB: " & (useWalkerAction.GetSizeOfAffectedTilesCount / 1024).ToString("0.000") & vbCrLf, MsgBoxStyle.Information)
    End Sub
    ''' <summary>
    ''' Download tiles within the radian @Latitude/Longitude with Zoom.
    ''' </summary>
    ''' <param name="radian">The radian.</param>
    Public Sub DownloadTiles(ByVal radian As Integer)
        If myTileProviders.GetCurrentURL = "" Then
            MsgBox("Set up the tile server in OSMtracker first.", MsgBoxStyle.Information)
        Else
            Dim x As Integer, y As Integer
            CalcTileXY(myLatitude, myLongitude, myZoom, x, y)
            Dim requiredPixel As Integer = CInt(radian * 1000 / MetersPerPixel(myZoom)) + 128
            Dim requiredTiles As Integer = requiredPixel \ 256 + 1
            If MsgBox("We have to download up to " & requiredTiles ^ 2 & " tiles. Continue?", MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
                Dim progressBar As New Progress
                progressBar.Show()
                Progress.BringToFront()
                progressBar.IncreaseDepth(CInt(requiredTiles ^ 2))
                Application.DoEvents()
                System.Windows.Forms.Cursor.Current = Cursors.WaitCursor
                Dim strFolder As String = myTileProviders.GetCurrentFolderName & IO.Path.DirectorySeparatorChar & myZoom
                If Not IO.Directory.Exists(strFolder) Then
                    Try
                        IO.Directory.CreateDirectory(strFolder)
                    Catch ex As Exception
                        MsgBox("Error creating folder: " & strFolder, MsgBoxStyle.Critical)
                        Exit Sub
                    End Try
                End If
                For i As Integer = x - requiredTiles + 1 To x + requiredTiles - 1
                    strFolder = myTileProviders.GetCurrentFolderName & IO.Path.DirectorySeparatorChar & myZoom & IO.Path.DirectorySeparatorChar & i
                    If Not IO.Directory.Exists(strFolder) Then
                        Try
                            IO.Directory.CreateDirectory(strFolder)
                        Catch ex As Exception
                            MsgBox("Error creating folder: " & strFolder, MsgBoxStyle.Critical)
                            Exit Sub
                        End Try
                    End If
                    For j As Integer = y - requiredTiles + 1 To y + requiredTiles - 1
                        Dim strFileName As String = strFolder & IO.Path.DirectorySeparatorChar & j & ".png"
                        If Not IO.File.Exists(strFileName) Then
                            Downloader.DownloadFile(myTileProviders.GetCurrentURL & myZoom & "/" & i & "/" & j & ".png", strFileName)
                        End If
                    Next
                    progressBar.Increase()
                    If progressBar.IsAborted Then
                        Exit For
                    End If
                Next
                progressBar.Close()
                System.Windows.Forms.Cursor.Current = Cursors.Default
                If progressBar.IsAborted Then
                    MsgBox("aborted.")
                Else
                    MsgBox("finished.")
                End If
            End If
        End If
    End Sub
    Private Function MetersPerPixel(ByVal zoom As Integer) As Double
        Const earthCircum As Double = 6378137 * 2 * Math.PI
        MetersPerPixel = earthCircum / ((1 << zoom) * 256)
    End Function

    Private powerof2 As Integer() = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, _
                                         2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576}


    ' Calculate the tile numbers for GPS position pos
    Private Sub CalcTileXY(ByVal lat As Double, ByVal lon As Double, ByVal zoom As Integer, ByRef x As Integer, ByRef y As Integer)
        Dim xf As Double = (lon + 180) / 360 * powerof2(zoom)
        Dim yf As Double = (1 - Math.Log(Math.Tan(lat * Math.PI / 180) + 1 / Math.Cos(lat * Math.PI / 180)) / Math.PI) / 2 * powerof2(zoom)
        x = CInt(Math.Floor(xf))
        y = CInt(Math.Floor(yf))
    End Sub
    Public ReadOnly Property TileProviderList() As List(Of String)
        Get
            Return myTileProviderList
        End Get
    End Property
    Public ReadOnly Property TileProviders() As TileProviders
        Get
            Return myTileProviders
        End Get
    End Property
    Public Property Zoom() As Integer
        Get
            Return myZoom
        End Get
        Set(ByVal value As Integer)
            myZoom = value
        End Set
    End Property
    Public Property Latitude() As Double
        Get
            Return myLatitude
        End Get
        Set(ByVal value As Double)
            myLatitude = value
        End Set
    End Property
    Public Property Longitude() As Double
        Get
            Return myLongitude
        End Get
        Set(ByVal value As Double)
            myLongitude = value
        End Set
    End Property
End Class
